unit Printer; {Unit by Doug Overmyer 11/91}

{********************  Interface   ************************}

Interface

uses WinTypes, WinProcs,WObjects,Strings;
const
	id_ChgPrnDlgOK       =  2599;
  id_ChgPrnDlgLB1      =  2598;

type
PAbortDlg = ^TAbortDlg;
TAbortDlg = object(TDialog)
	procedure Cancel(var Msg:TMessage);virtual id_First+id_Cancel;
end;

type
PChgPrnDlg = ^TChgPrnDlg;      {used by ChangePrinter()}
TChgPrnDlg = object(TDialog)
	AllDevicesBuf:Array[0..4096] of Char;
  procedure WMInitDialog(var Msg:TMessage);virtual wm_First+wm_InitDialog;
	procedure IDChgPrtDlgOK(var Msg:TMessage);virtual id_First+id_ChgPrnDlgOK;
end;

type
PPrinter = ^TPrinter;
TPrinter = object(TObject)
	PrintDC:HDC;
  CurFont:HFont;
  LineX,LineY:Integer; {line width and height for CurFont}
  PageNumber:Integer;  {current page number}
  Metrics:TTextMetric; {current text metrics}
  DeviceName:Array[0..79] of Char;  {win.ini device= }
  DriverName:Array[0..79] of Char;
  OutPort:Array[0..79] of Char;
  EscResult:Integer;  {result of most recent ESC call}
  CurX,CurY:Integer;  {print Cursor X,Y position in device units}
  PageX,PageY:Integer;{page width and height in device units}
  LogPixelsX,LogPixelsY:Integer; {pixels/inch}
  Printing:Boolean;
  constructor Init;
  destructor Done;virtual;
  function GetPrintDC:Boolean;virtual;
  procedure GetPrnParms;virtual;
  procedure InstallAbortProc;virtual;
  procedure SetupPage;virtual;
	function PrnStart(DocName:PChar):Boolean;virtual;
	procedure NewPage;virtual;
	procedure PrnStop;virtual;
  function GetIC:HDC;virtual;
  function  DeleteIC:Boolean;virtual;
  function SetFont(NewFont:HFont):HFont;virtual;
  procedure PrnDeviceMode(ParentWindow:PWindowsObject);virtual;
  procedure ChangePrinter(ParentWindow:PwindowsObject);virtual;
end;

TDeviceMode = function(HWindow:HWnd;Module:THandle;
		DevName,OutPortName:PChar):Boolean;
TExtDeviceMode = function(HWindow:HWnd;
		HDriver:THandle;
    OutMode:TDevMode;
    DevName:PChar;
    OutPut:PChar;
    InMode:TDevMode;
    Profile:PChar;
    Mode:Word):Integer;

PLPrinter = ^TLPrinter;
TLPrinter = object(TPrinter)
	Margin:TRect;  {margin in device units}
  FooterY:Integer; {height of footer in device units}
  IsFooter:Boolean;  {True = is printing footer}
  constructor Init;
  procedure PrintLine(aString:PChar);virtual;
  procedure PrnStop;virtual;
  procedure CheckNewPage;virtual;
  procedure DoHeader;virtual;
  procedure DoFooter;virtual;
  procedure SetupPage;virtual;
  procedure SetTOP;virtual;
  procedure SetMarginL(NewMargin:Integer);virtual;
  procedure SetMarginR(NewMargin:Integer);virtual;
  procedure SetMarginT(NewMargin:Integer);virtual;
  procedure SetMarginB(NewMargin:Integer);virtual;
end;
{to use a TLPrinter object, call
	0. Subclass DoHeader,DoFooter properly
	1. New.....Init
  2. PrtStart
  3. SetMarginL, SetFont, etc. Also set FooterY as needed.
  4. SetupPage  (does the DoHeader if any)
  5. printLine  (as needed)
  6. DoFooter, PrnStop (do one last footer before stopping)
  7. SetFont (if needed to restore 'old font')
  8. Dispose
alternative - to get a PrintDC only
	1. New...Init
  2. GetIC
  3.  use the IC, TextMetrics, other info, but NO drawing/printing
  4. DeleteIC
  5. Dispose
}
{**********************  Implementation    *********************}
Implementation
{$R printer.res}
{*********************    Globals          *********************}
var
	Aborted:Boolean;
  AbortDlg:PAbortDlg;
  PAbortProc:TFarProc;

function StrTok(P:PChar;C:Char):PChar;
const
	Next:Pchar = nil;
begin
	if P = NIL then P := Next;
  Next := StrScan(P,C);
  If Next <> NIL then
  	begin
    Next^ := #0;
    Next := Next+1;
  end;
  StrTok := P;
end;

function AbortProc(PrtDC:HDC;Code:Integer):Boolean;export;
var
	Msg:TMsg;
begin
	while (not Aborted) and PeekMessage(Msg,0,0,0,pm_Remove) do
  	if not IsDialogMessage(AbortDlg^.HWindow,Msg) then
    	begin
      translateMessage(Msg);
      DispatchMessage(Msg);
      end;
  AbortProc := not Aborted;
end;

{***************************************************************}
constructor TPrinter.Init;
begin
	TObject.Init;
  PrintDC := 0;
  CurX := 0;
  CurY := 0;
  Aborted := False;
  CurFont := GetStockObject(Device_Default_Font);
  PageNumber := 1;
end;

destructor TPrinter.Done;
begin
	TObject.Done;
end;

function TPrinter.GetPrintDC;
var
	Buf1 :Array[0..80] of Char;
begin
	GetProfileString('Windows','device',',,',Buf1,SizeOf(Buf1));
  StrCopy(DeviceName,StrTok(Buf1,','));
  StrCopy(DriverName,StrTok(nil,','));
  Strcopy(OutPort,StrTok(nil,','));
  PrintDC := CreateDc(DriverName,DeviceName,OutPort,nil);
  GetPrintDC := (PrintDC > 0);
end;

procedure TPrinter.GetPrnParms;
begin
	GetTextMetrics(PrintDC,Metrics);
  LogPixelsX := GetDeviceCaps(PrintDC,LogPixelsX);
  LogPixelsY := GetDeviceCaps(PrintDC,LogPixelsY);
  PageX := GetDeviceCaps(PrintDC,HorzRes);
  PageY := GetDeviceCaps(PrintDC,VertRes);
  LineY := Metrics.tmHeight + Metrics.tmExternalLeading;
end;

procedure TPrinter.InstallAbortProc;
begin
	AbortDlg := new(PAbortDlg,Init(Application^.MainWindow,
  		'AbortDlg'));
  AbortDlg^.EnableAutoCreate;
  Application^.MakeWindow(AbortDlg);
	PAbortProc := MakeProcInstance(@AbortProc,HInstance);
  EscResult := Escape(PrintDC,SetAbortProc,0,PAbortProc,nil);
end;

procedure TPrinter.SetupPage;
begin
	{Formal method}
end;

function TPrinter.PrnStart(DocName:PChar):Boolean;
begin
	GetPrintDC;
  If PrintDC <> 0 then
  	begin
    GetPrnParms;
    InstallAbortProc;
    if EscResult > 0 then
    	begin
    	EscResult := Escape(PrintDC,StartDoc,StrLen(Docname),DocName,nil);
    	Printing := (EscResult > 0);
      end;
    end
  else
  	Printing := false;
  If not Printing then
  	begin
    if AbortDlg <> nil then
    		AbortDlg^.CloseWindow;
    MessageBox(Application^.MainWindow^.HWindow,'Printer Initialization Failed',
    	'Error',mb_IconExclamation or mb_OK);
    end;
  PrnStart := Printing;
end;

procedure TPrinter.NewPage;
begin
	if Printing and (EscResult > 0) then
  	EscResult := Escape(PrintDC,NewFrame,0,nil,nil);
  SelectObject(PrintDC,CurFont);
  Inc(PageNumber);
end;

procedure TPrinter.PrnStop;
begin
	if Printing then
  	begin
    if AbortDlg <> nil then
    	AbortDlg^.CloseWindow;
  	if EscResult > 0 then
    	Escape(PrintDC,EndDoc,0,nil,nil);
    DeleteDC(PrintDC);
    Printing := false;
    end;
end;

function TPrinter.GetIC:HDC;
var
	Buf1 :Array[0..80] of Char;
begin
	GetProfileString('Windows','device',',,',Buf1,SizeOf(Buf1));
  StrCopy(DeviceName,StrTok(Buf1,','));
  StrCopy(DriverName,StrTok(nil,','));
  Strcopy(OutPort,StrTok(nil,','));
  PrintDC := CreateIC(DriverName,DeviceName,OutPort,nil);
  GetPrnParms;
  GetIC:= PrintDC;
end;

function TPrinter.DeleteIC;
begin
	DeleteIC := (PrintDC > 0);
	if PrintDC > 0 then
    DeleteDC(PrintDC);
end;

function TPrinter.SetFont(NewFont:HFont):HFont;
begin
	SetFont :=SelectObject(PrintDC,NewFont);
  CurFont := NewFont;
  GetPrnParms;
end;

procedure TPrinter.PrnDeviceMode(ParentWindow:PWindowsObject);
var
	ExtDeviceMode:TExtDeviceMode;
  DeviceMode:TDeviceMode;
  FullDrvName:Array[0..12] of Char;
  HDriver:THandle;
  PA1,PA2:TFarProc;
  DMode:TDevMode;
begin
	StrCat(StrCopy(FullDrvName,DriverName),'.DRV');
  HDriver := LoadLibrary(FullDrvName);
  If HDriver < 32 then
  	begin
    MessageBox(ParentWindow^.HWindow,'Failed to load Printer driver',
    	'Check Printer setup',mb_IconExclamation or mb_OK);
    Exit;
    end;
  PA1 := GetProcAddress(HDriver,'ExtDeviceMode');
  PA2 := GetProcAddress(HDriver,'DeviceMode');
  If PA1 <> nil then
  	begin
    ExtDeviceMode := TExtDeviceMode(PA1);
    ExtDeviceMode(Application^.MainWindow^.HWindow,
    		HDriver,
        DMode,
        DeviceName,
        OutPort,
        DMode,
        nil,
        dm_prompt or dm_Update);
    end
  else if PA2 <> nil then
  	begin
    DeviceMode := TDeviceMode(PA2);
    DeviceMode(Application^.MainWindow^.HWindow,
    	HDriver,
      DeviceName,
      OutPort);
    end;
  FreeLibrary(HDriver);
  SetFocus(ParentWindow^.HWindow);
end;

procedure TPrinter.ChangePrinter(ParentWindow:PWindowsObject);
var
	ChgPrnDlg:PChgPrnDlg;
begin
	ChgPrnDlg := New(PChgPrnDlg,Init(Parentwindow,'ChgPrnDlg'));
  Application^.ExecDialog(ChgPrnDlg);
end;

{*********************  TAbortDlg  **********************}
procedure TAbortDlg.Cancel(var Msg:TMessage);
begin
	Aborted := True;
  TDialog.Cancel(Msg);
end;

{********************  TChgPrnDlg   ************************}
procedure TChgPrnDlg.WMInitDialog(var Msg:TMessage);
var
 	pAllDevicesBuf:PChar;
 	Buf:Array[0..64] of Char;
 	pBuf:PChar;
 	Printer1:Array[0..64] of Char;
 	Printer:Array[0..64] of Char;
	pPrinter:PChar;
begin
	GetProfileString('devices',nil,'',AllDevicesBuf,sizeof(AllDevicesBuf));
	TDialog.WMInitDialog(Msg);
  pAllDevicesBuf := AllDevicesBuf;
  pBuf := @Buf;
  pPrinter := @Printer;
  repeat
    StrCopy(Buf,pAllDevicesBuf);
    GetProfileString('devices',Buf,'',Printer1,sizeof(Printer1));
    StrCat(StrCat(StrCopy(Printer,Buf),','),Printer1);
  	SendDlgItemMsg(id_ChgPrnDlgLB1,lb_AddString,word(0),LongInt(pPrinter));
    pAllDevicesbuf := pAllDevicesBuf+StrLen(pBuf)+1;
  until StrLen(pAllDevicesBuf) = 0;
end;

procedure TChgPrnDlg.IDChgPrtDlgOK(var Msg:TMessage);
var
	Idx:Integer;
  Buf:Array[0..64] of Char;
  Ptr:PChar;
  NewDevice:Array[0..64] of Char;
begin
	StrCopy(Buf,'');
	Ptr := @Buf;
  Idx := SendDlgItemMsg(id_ChgPrnDlgLB1,lb_GetCurSel,0,0);
  if Idx <> lb_Err then
  	SendDlgItemMsg(id_ChgPrnDlgLB1,lb_GetText,idx,Longint(Ptr));
  if StrLen(Ptr) > 0 then
  	begin
    StrCopy(NewDevice,Buf);
    WriteProfileString('Windows','device',NewDevice);
    end;
	EndDlg(1);
end;

{**********************  TLPrinter    ************************}
constructor TLPrinter.Init;
begin
	TPrinter.Init;
  Margin.Left := 0;
  Margin.Right := 0;
  Margin.Top := 0;
  Margin.Bottom := 0;
  FooterY := 0;
  IsFooter := False;
end;

procedure TLPrinter.PrintLine(aString:PChar);
begin
  CurX := Margin.Left;
	TextOut(PrintDC,CurX,CurY,aString,StrLen(aString));
  CurY := CurY + LineY;
  CurX := Margin.Left;
  if not IsFooter then
  	CheckNewPage;
end;

procedure TLPrinter.PrnStop;
begin
	if CurY > Margin.Top then
  	NewPage;
  TPrinter.PrnStop;
end;

procedure TLPrinter.CheckNewPage;
begin
	if (CurY + Margin.Bottom + 2*LineY + FooterY ) > PageY then
  	begin
    IsFooter := True;
  	DoFooter;
    IsFooter := False;
    NewPage;
    SetupPage;
    end;
end;

procedure TLPrinter.DoHeader;
begin
	{formal method}
end;

procedure TLPrinter.DoFooter;
begin
	IsFooter := True;
  {Include this code when you subclass DoFooter}
  IsFooter := False;
end;

procedure TLPrinter.SetupPage;
begin
	SetTOP;
  DoHeader;
end;

procedure TLPrinter.SetTOP;
begin
	CurX := Margin.Left;
  CurY := Margin.Top;
end;

procedure TLPrinter.SetMarginL(NewMargin:Integer);
begin
	Margin.Left := NewMargin;
end;

procedure TLPrinter.SetMarginR(NewMargin:Integer);
begin
	Margin.Right := NewMargin;
end;

procedure TLPrinter.SetMarginT(NewMargin:Integer);
begin
	Margin.Top := NewMargin;
end;

procedure TLPrinter.SetMarginB(NewMargin:Integer);
begin
	Margin.Bottom := NewMargin;
end;
{***************************************************************}
end.